package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.espec.ESpecLogic;
import be.iminds.ilabt.jfed.espec.model.*;
import be.iminds.ilabt.jfed.espec.parser.ESpecConstants;
import be.iminds.ilabt.jfed.espec.parser.ESpecDefaults;
import be.iminds.ilabt.jfed.espec.util.ChModPermissionsUtil;
import be.iminds.ilabt.jfed.espec.util.ESpecLogger;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLog;
import be.iminds.ilabt.jfed.espec.util.LimitedLiveLogImpl;
import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.InputStreamToLogsThread;
import be.iminds.ilabt.jfed.highlevel.util.LogEntryGeneratorWrappingLogger;
import be.iminds.ilabt.jfed.highlevel.util.StringMemorySourceFile;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.rspec.basic_model.BasicStringRspec;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.util.lib.AnsibleFileWriter;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.ConnectionException;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.sftp.SFTPClient;
import net.schmizz.sshj.transport.TransportException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.slf4j.event.Level;

import javax.annotation.Nonnull;
import java.io.IOException;
import java.security.KeyPair;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

/**
 * User: twalcari
 * Date: 5/27/2016
 * Time: 11:51
 */
public class ExperimentSpecificationServicesAnsibleState extends SlicedState<ExperimentSpecificationServicesAnsibleState.ExperimentSpecificationServicesAnsibleStateSlice> {
    private static final Logger ACTUAL_LOG = LoggerFactory.getLogger(ExperimentSpecificationServicesAnsibleState.class);
    private final Logger LOG;

    @Nonnull private final SetupSoftwareExperimentJob job;
    @Nonnull private final Experiment experiment;
    @Nonnull private final ExperimentSpecification experimentSpecification;
    @Nonnull private final GeniUserProvider geniUserProvider;
    @Nonnull private final RemotePathResolver remotePathResolver;
    @Nonnull private final AnsibleSpec ansibleSpec;

    @Nonnull private final ESpecLogger eSpecLogger;

    private final boolean useRemoteAnsibleNode;
    @Nonnull private final String ansibleNodeClientId;
    @Nonnull private final RspecNode ansibleNode;


    @Nonnull private final List<ExperimentSpecificationServicesAnsibleStateSlice> slices;

    protected ExperimentSpecificationServicesAnsibleState(@Nonnull SetupSoftwareExperimentJob job,
                                                          @Nonnull GeniUserProvider geniUserProvider,
                                                          @Nonnull RemotePathResolver remotePathResolver,
                                                          @Nonnull AnsibleSpec ansibleSpec) {
        super("Execute ansible services");
        this.job = job;
        this.LOG = new LogEntryGeneratorWrappingLogger(ACTUAL_LOG, job.getJobReport());
        this.geniUserProvider = geniUserProvider;
        this.remotePathResolver = remotePathResolver;
        this.ansibleSpec = ansibleSpec;
        this.experiment = job.getExperiment();
        if (job.getExperiment().getExperimentSpecification() == null)
            throw new IllegalStateException("job.getExperiment().getExperimentSpecification() is null");
        this.experimentSpecification = job.getExperiment().getExperimentSpecification();
        this.eSpecLogger = experiment.getExperimentSpecificationLogger();
        experiment.requireSlice();

        /////////////

        assert ESpecLogic.hasAnsibleNode(experimentSpecification);
        this.useRemoteAnsibleNode = ansibleSpec.getAnsibleHostSpec().getHostType() != AnsibleHostSpec.HostType.LOCAL;

        if (!useRemoteAnsibleNode) {
            throw new RuntimeException("LOCAL ansible execute is not yet implemented.");
        } else {
            if (ansibleSpec.getAnsibleHostSpec().getNodeName() == null)
                throw new IllegalStateException("ansible node name should be set");
        }

        assert experiment.getNewRequestRspecSource() != null;
        ModelRspec request = experiment.getNewRequestRspecSource() == null ?
                null :
                experiment.getNewRequestRspecSource().getImmutableModelRspec();

        if (request == null) {
            String err = "Cannot find ansible node because parsing request rspec failed";
            LOG.error(err);
            throw new RuntimeException(err);
        }

        RspecNode ansibleNode = request.getNodeByClientId(ansibleSpec.getAnsibleHostSpec().getNodeName());
        if (ansibleNode == null || ansibleNode.getClientId() == null) {
            String err = "Cannot find ansible node \""+ansibleSpec.getAnsibleHostSpec().getNodeName()+"\" in request rspec";
            LOG.error(err);
            throw new RuntimeException(err);
        }
        this.ansibleNode = ansibleNode;
        this.ansibleNodeClientId = ansibleNode.getClientId();

        this.slices = Collections.singletonList(
                new ExperimentSpecificationServicesAnsibleStateSlice(job, ansibleNode, ansibleSpec)
        );
    }

    @Override
    public Collection<ExperimentSpecificationServicesAnsibleStateSlice> getSlices() {
        return slices;
    }

    public class ExperimentSpecificationServicesAnsibleStateSlice extends StateSlice {
        @Nonnull  protected final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
        @Nonnull protected final RspecNode ansibleNode;
        @Nonnull protected final AnsibleSpec ansibleSpec;

        @Nonnull private String galaxyCommand;
        @Nonnull private String playbookCommand;

        //only initialised when there is an SSH connection (relies on homedir location)
        private String ansibleDirPath;

        public ExperimentSpecificationServicesAnsibleStateSlice(@Nonnull SetupSoftwareExperimentJob job,
                                                                @Nonnull RspecNode ansibleNode,
                                                                @Nonnull AnsibleSpec ansibleSpec) {
            super(job);
            this.setupSoftwareExperimentJob = job;
            this.ansibleNode = ansibleNode;
            this.ansibleSpec = ansibleSpec;

            this.galaxyCommand = ansibleSpec.getAnsibleHostSpec().getGalaxyCommand() == null ?
                    ESpecDefaults.ANSIBLE_GALAXY_COMMAND :
                    ansibleSpec.getAnsibleHostSpec().getGalaxyCommand();
            this.playbookCommand = ansibleSpec.getAnsibleHostSpec().getPlaybookCommand() == null ?
                    ESpecDefaults.ANSIBLE_PLAYBOOK_COMMAND :
                    ansibleSpec.getAnsibleHostSpec().getPlaybookCommand();
        }

        @Override
        public String getName() {
            return "Ansible on " + ansibleNode.getClientId();
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            SSHClient ssh;
            try {
                ssh = setupSoftwareExperimentJob.getSSHClient(ansibleNode, this);
            } catch (SetupSoftwareExperimentJob.SshException e) {
                LOG.warn("Aborting ExecuteAnsibleServiceStateSlice.statefulRun due to SSH comnection problem", e);
                return ExperimentTaskStatus.FAILED;
            }
            assert ssh != null;
            ExternalFileUtil externalFileUtil = new ExternalFileUtil(ssh);

            String ansibleDirContentPath = remotePathResolver.getRemotePath(ESpecConstants.DirContent.ANSIBLE, ansibleNodeClientId);
            if (ansibleDirContentPath == null)
                throw new IllegalStateException("Unknown ansible content dir");
            this.ansibleDirPath = ensureEndsWithSlash(ansibleDirContentPath);

            eSpecLogger.firePreAnsible();

            PrepareAnsibleDirState prepareAnsibleDirState = new PrepareAnsibleDirState(ssh);
            setAndRunState(prepareAnsibleDirState);

            ExperimentTaskStatus worstStatus = prepareAnsibleDirState.getStatus();

            for (AnsibleGalaxySpec gs : ansibleSpec.getAnsibleGalaxySpecs()) {
                RunGalaxyState executeGalaxyState = new RunGalaxyState(gs, ssh);
                setAndRunState(executeGalaxyState);
                worstStatus = ExperimentTaskStatus.worstOf(worstStatus, executeGalaxyState.getStatus());
            }

            for (AnsiblePlaybookSpec pbs : ansibleSpec.getAnsiblePlaybookSpecs()) {
                ExecutePlayBookState executePlayBookState = new ExecutePlayBookState(pbs, ssh);
                setAndRunState(executePlayBookState);
                worstStatus = ExperimentTaskStatus.worstOf(worstStatus, executePlayBookState.getStatus());
            }

            setupSoftwareExperimentJob.closeSSH(ansibleNode);

            eSpecLogger.firePostAnsibleAll(worstStatus == ExperimentTaskStatus.SUCCESS);

            return worstStatus;
        }

        public class PrepareAnsibleDirState extends State {
            @Nonnull private final SSHClient ssh;

            protected PrepareAnsibleDirState(@Nonnull SSHClient ssh) {
                super("Preparing ansible dir '"+ansibleDirPath+"'");

                this.ssh = ssh;

                if (experiment.getSlice().getManifestRspec() == null)
                    throw new IllegalStateException("Cannot prepare ansible node because there is no manifest RSpec.");
                String manifestString = experiment.getSlice().getManifestRspec().getRspecXmlString();
                if (manifestString == null || manifestString.isEmpty())
                    throw new IllegalStateException("Cannot prepare ansible node because there is an empty manifest RSpec.");
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                KeyPair ansibleKeyPair = experiment.getKeypairs().values().iterator().next();

                if (experiment.getSlice().getManifestRspec() == null)
                    throw new IllegalStateException("Cannot prepare ansible node because there is no manifest RSpec.");
                String manifestString = experiment.getSlice().getManifestRspec().getRspecXmlString();
                if (manifestString == null || manifestString.isEmpty())
                    throw new IllegalStateException("Cannot prepare ansible node because there is an empty manifest RSpec.");

                AnsibleFileWriter ansibleFileWriter = AnsibleFileWriter.createWithCopiedPrivateKey(
                        new BasicStringRspec(manifestString),
                        ansibleKeyPair.getPrivate(),
                        ansibleKeyPair.getPublic(),
                        geniUserProvider.getLoggedInGeniUser(),
                        null /* this is meant for testbed node, so it never needs any proxy */,
                        geniUserProvider.getLoggedInGeniUser().getUserUrn().getResourceName());
                //ansibleFileWriter.addExcludedClientId(ansibleNodeClientId);

                if (ansibleSpec.getGroups() != null)
                    ansibleFileWriter.addExtraGroups(ansibleSpec.getGroups());

                List<AnsibleFileWriter.AnsibleFileInfo> ansibleFiles = ansibleFileWriter.getFilesToWrite(false);

                try (SFTPClient sftpClient = ssh.newSFTPClient()) {
                    for (AnsibleFileWriter.AnsibleFileInfo ansibleFileInfo : ansibleFiles) {
                        String fullPath = ansibleDirPath + "" + ansibleFileInfo.getFilename();
                        try {
                            ssh.newSCPFileTransfer().upload(
                                    new StringMemorySourceFile(ansibleFileInfo.getFilename(), ansibleFileInfo.getContent()),
                                    fullPath);

                            if (ansibleFileInfo.isPrivateFile()) {
                                //NOTE: sftpClient.chmod and other permsision related commands/parameters in sshj SFTP do not work reliably...
                                sftpClient.chmod(fullPath, ChModPermissionsUtil.toMask("u=r"));

                                //So we also try another way as backup
//                            final FileAttributes attrs = sftpClient.statExistence(fullPath);
//                            if (attrs.getPermissions() ?) {
                                ssh.startSession().exec("chmod u=r " + fullPath);
                            }
                        } catch (IOException ex) {
                            LOG.error("Error while uploading inventory-file to {} on {}", fullPath, ansibleNodeClientId, ex);
                            updateMessage(String.format("Error while uploading inventory-file to %s on %s: %s", fullPath, ansibleNodeClientId, ex.getMessage()));
                            return ExperimentTaskStatus.FAILED;
                        }
                    }
                } catch (IOException ex) {
                    LOG.error("Error while initialising SFTP client for ansible preparations", ex);
                    updateMessage(String.format("Error while initialising SFTP client for ansible preparations: %s", ex.getMessage()));
                    return ExperimentTaskStatus.FAILED;
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class RunGalaxyState extends State {
            @Nonnull private final AnsibleGalaxySpec galaxySpec;
            @Nonnull private final SSHClient ssh;

            protected RunGalaxyState(@Nonnull AnsibleGalaxySpec galaxySpec, @Nonnull SSHClient ssh) {
                super("Running galaxy requirements '" + galaxySpec.getDesc() + "'");

                this.galaxySpec = galaxySpec;
                this.ssh = ssh;
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                try {
                    String remotePath = remotePathResolver.getRemotePath(galaxySpec, ansibleNodeClientId);
                    if (remotePath == null) {
                        String err = "Did not find remote path of ansible galaxy requirements. Bug in ansible step?";
                        LOG.error(err);
                        updateMessage(err);
                        throw new RuntimeException(err);
                        //                    return ExperimentTaskStatus.FAILED;
                    }

                    LimitedLiveLog liveLog = new LimitedLiveLogImpl();
                    //TODO: log ansible galaxy stuff as well!

                    String logPath;
                    if (galaxySpec.getLog() == null) {
                        if (experimentSpecification.getConfig().isStoreRemoteLogsByDefault())
                            logPath = ESpecDefaults.deriveAutomaticGalaxyLogFile(remotePath);
                        else
                            logPath = null;
                    } else {
                        logPath = remotePathResolver.getRemoteLogPath(galaxySpec, ansibleNodeClientId);
                    }

                    //install ansible galaxy requirements
                    String galaxyCommandString = String.format(
                            "%s install -r %s -p roles", galaxyCommand, remotePath);
                    LOG.debug("Executing galaxy command: "+galaxyCommandString);

                    String command = ExperimentSpecificationServicesSingleExecuteState.createSshCommand(
                            galaxyCommandString,
                            ansibleDirPath,
                            logPath,
                            false,
                            experimentSpecification
                    );

                    LOG.debug("Actual SSH exec command: "+command);

                    Session.Command c =
                            ssh.startSession().exec(command);

                    new InputStreamToLogsThread(c.getInputStream(),
                            LOG, Level.INFO, "GALAXY-OUTPUT: ",
                            liveLog).start();
                    new InputStreamToLogsThread(c.getErrorStream(),
                             LOG, Level.WARN, "GALAXY-ERROR: ",
                            liveLog).start();

                    c.join();

                    if (c.getExitStatus() != 0) {
                        updateMessage("ansible-galaxy execution was unsuccessful. Aborting");
                        return ExperimentTaskStatus.FAILED;
                    }
                } catch (TransportException | ConnectionException ex) {
                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
                }

                return ExperimentTaskStatus.SUCCESS;
            }
        }

        public class ExecutePlayBookState extends State {
            @Nonnull private final AnsiblePlaybookSpec playbook;
            @Nonnull private final SSHClient ssh;

            protected ExecutePlayBookState(@Nonnull AnsiblePlaybookSpec playbook, @Nonnull SSHClient ssh) {
                super("Executing ansible-playbook '" + playbook.getDesc() + "'");

                this.playbook = playbook;
                this.ssh = ssh;
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                String remotePath = remotePathResolver.getRemotePath(playbook, ansibleNodeClientId);
                long logId = eSpecLogger.getNextLogId();
                if (remotePath == null) {
                    String err = "Did not find remote path of ansible playbook. Bug in ansible step?";
                    LOG.error(err);
                    updateMessage(err);
                    eSpecLogger.firePostAnsiblePlaybookFail(logId, playbook, "not found", null,
                            ansibleNodeClientId, err, null);
                    throw new RuntimeException(err);
                    //                    return ExperimentTaskStatus.FAILED;
                }

                LimitedLiveLog liveLog = new LimitedLiveLogImpl();
                eSpecLogger.firePreAnsiblePlaybook(logId, playbook, "not found", null,
                        ansibleNodeClientId, liveLog);

                String logPath;
                if (playbook.getLog() == null) {
                    if (experimentSpecification.getConfig().isStoreRemoteLogsByDefault())
                        logPath = ESpecDefaults.deriveAutomaticPlaybookLogFile(remotePath);
                    else
                        logPath = null;
                } else {
                    logPath = remotePathResolver.getRemoteLogPath(playbook, ansibleNodeClientId);
                }

                try {
                    String arguments = "";
                    arguments += remotePath;

                    //ansible debug uses one to 3 v's  -v  -vv  or  -vvv
                    if (playbook.isDebug()) {
                        arguments += " -";
                        for (int i = 0; i < playbook.getDebug(); i++)
                            arguments += "v";
                    }

                    if (playbook.hasExtraVarsString()) {
                        arguments += " --extra-vars '"+playbook.getExtraVarsString()+"'";
                    }
                    if (playbook.hasExtraVarsJsonFileSpec()) {
                        AnsiblePlaybookSpec.ExtraVarsJsonFileSpec extraVarsJson =
                                playbook.getExtraVarsJsonFileSpec();
                        if (extraVarsJson == null)
                            throw new IllegalStateException("Illegal state of playbook related to extravars");

                        //find real path of the extravars file
                        String extraVarsRemotePath =
                                remotePathResolver.getRemotePath(extraVarsJson, ansibleNodeClientId);
                        arguments += " --extra-vars '@"+ extraVarsRemotePath+ "'";
                    }

                    //don't need this if AnsibleFileWriter writes the ansible files and uses ansible config
//                    arguments += " -i '"+ansibleInventoryPath+"'";

                    //don't need this if AnsibleFileWriter writes the ansible files
//                    arguments += " -e 'host_key_checking=False'";

                    String playbookCommandString = playbookCommand + " " + arguments;

                    String command = ExperimentSpecificationServicesSingleExecuteState.createSshCommand(
                            playbookCommandString,
                            ansibleDirPath,
                            logPath,
                            false,
                            experimentSpecification
                    );

                    LOG.debug("Executing playbook command: "+playbookCommandString);
                    LOG.debug("Actual SSH exec command: "+command);
                    Session.Command c =
                            ssh.startSession().exec(command);

                    new InputStreamToLogsThread(c.getInputStream(),
                            LOG, Level.INFO, "PLAYBOOK-OUTPUT: ",
                            liveLog).start();
                    new InputStreamToLogsThread(c.getErrorStream(),
                            LOG, Level.WARN, "PLAYBOOK-ERROR: ",
                            liveLog).start();

                    c.join();

                    if (c.getExitStatus() != 0) {
                        updateMessage("ansible-playbook execution was unsuccessful.");
                        eSpecLogger.firePostAnsiblePlaybookFail(logId, playbook, remotePath, logPath, ansibleNodeClientId,
                                "ansible-playbook execution was unsuccessful. exit status = "+c.getExitStatus(), null);
                        return ExperimentTaskStatus.FAILED;
                    } else {
                        eSpecLogger.firePostAnsiblePlaybookSuccess(logId, playbook, remotePath, logPath, ansibleNodeClientId);
                        return ExperimentTaskStatus.SUCCESS;
                    }
                } catch (TransportException | ConnectionException ex) {
                    eSpecLogger.firePostAnsiblePlaybookFail(logId, playbook, remotePath, logPath, ansibleNodeClientId,
                            "SSH-exception while executing ansible", ex);
                    throw new JFedException("SSH-exception while executing ansible: " + ex.getMessage(), ex);
                }
            }
        }
    }

    @Nonnull
    private static String ensureEndsWithSlash(@Nonnull String p) {
        if (p.endsWith("/")) return p;
        return p + "/";
    }

    @Nonnull
    private static String ensureStartsWithSlash(@Nonnull String p) {
        if (p.startsWith("/")) return p;
        return "/" + p;
    }
}
